#!/usr/bin/env node
import fs from 'fs';
import path from 'path';
import process from 'process';
import browserSync from 'browser-sync';
import chalk from 'chalk';
import chokidar from 'chokidar';
import commandLineArgs from 'command-line-args';
import merge from 'deepmerge';
import { isMarkdownFile } from './utilities/files.js';
import Doxicity, { defaultConfig } from './index.js';
const bs = browserSync.create();
let options;
let config = {
    ...defaultConfig
};
try {
    options = {
        dir: process.cwd(),
        init: false,
        port: 8080,
        serve: false,
        watch: false,
        ...commandLineArgs([
            { name: 'dir', alias: 'd', type: String },
            { name: 'init', type: Boolean },
            { name: 'port', alias: 'p', type: Number },
            { name: 'serve', alias: 's', type: Boolean },
            { name: 'watch', alias: 'w', type: Boolean }
        ])
    };
}
catch (err) {
    console.error(chalk.red(err.message));
    process.exit(1);
}
const configFile = path.resolve(path.join(options.dir, 'doxicity.config.js'));
let userConfig = {};
if (options.init) {
    if (fs.existsSync(configFile)) {
        console.error(`${chalk.red(`A config file cannot be created here because one already exists!\n\n`)}${configFile}`);
        process.exit(1);
    }
    try {
        const template = `export default {
  source: '.',
  destination: './_site',
  baseUrl: '',
  layouts: 'layouts',
  partials: 'partials',
  assets: '',
  data: {},
  cleanBeforePublishing: false,
  plugins: [],
  helpers: [],
  pageExtension: 'html'
};
`;
        fs.writeFileSync(configFile, template, 'utf-8');
    }
    catch (err) {
        console.error(`${chalk.red(`Failed to create a config file at the following location.\n\n`)}${configFile}\n\n` +
            `Please make sure Doxicity has write access to the target directory and try again.`);
        process.exit(1);
    }
}
try {
    userConfig = (await import(configFile)).default;
    config = merge(defaultConfig, userConfig);
    process.chdir(path.resolve(options.dir));
}
catch {
    console.error(`${chalk.red(`Unable to locate a config file in the following directory:\n\n`)}${path.resolve(options.dir)}\n\n` +
        `Please make sure your config file is named "doxicity.config.js" and try again. ` +
        `Tip: you can quickly generate a config file using the "doxicity --init" command.`);
    process.exit(2);
}
config.source = path.resolve(config.source);
config.destination = path.resolve(config.destination);
const dox = new Doxicity(config);
try {
    await dox.publish();
}
catch (err) {
    console.error(`${chalk.red(`An publishing error occurred!`)}\n\n${err.message}`);
    process.exit(3);
}
if (options.watch) {
    console.log(`Watching for changes in: ${config.source}`);
    const pathsToWatch = [path.join(config.source, '**/*'), configFile];
    let isConfigDirty = false;
    const watcher = chokidar.watch(pathsToWatch, {
        persistent: true,
        ignored: [path.resolve(config.destination)],
        ignoreInitial: true
    });
    watcher
        .on('add', async (filename) => {
        if (isMarkdownFile(filename)) {
            console.log(`New page created: "${filename}"`);
            await dox.publish({
                copyAssets: false
            });
        }
        else {
            console.log(`New file created: "${filename}"`);
            await dox.publish();
        }
        if (options.serve) {
            bs.reload();
        }
    })
        .on('change', async (filename) => {
        if (filename === configFile) {
            isConfigDirty = true;
        }
        if (isConfigDirty) {
            console.log(chalk.yellow(`** Your config file has changed! Please restart the dev server for your changes to take effect. **`));
        }
        if (isMarkdownFile(filename)) {
            console.log(`Page changed: "${filename}"`);
            await dox.publish({ copyAssets: false });
        }
        else {
            console.log(`File changed: "${filename}"`);
            await dox.publish();
        }
        if (options.serve) {
            bs.reload();
        }
    })
        .on('unlink', async (filename) => {
        console.log(`Page removed: "${filename}"`);
        try {
            await dox.clean();
        }
        catch (err) {
            console.error(chalk.yellow(`Unable to clean the output directory before publishing: ${err.message}`));
        }
        try {
            await dox.publish();
        }
        catch (err) {
            console.error(chalk.red(err.message));
        }
        if (options.serve) {
            bs.reload();
        }
    });
}
if (options.serve) {
    bs.init({
        open: true,
        startPath: '/',
        port: options.port,
        logLevel: 'silent',
        logPrefix: '[doxicity]',
        logFileChanges: false,
        notify: false,
        ghostMode: false,
        server: {
            baseDir: config.destination
        }
    }, () => {
        const url = `http://localhost:${options.port}`;
        console.log(chalk.cyan(`Launched the Doxicity dev server at ${url} 📗`));
    });
}
