import fs from 'fs/promises';
import path from 'path';
import browserSync from 'browser-sync';
import chalk from 'chalk';
import chokidar from 'chokidar';
import cpy from 'cpy';
import merge from 'deepmerge';
import { globby } from 'globby';
import Handlebars from 'handlebars';
import { JSDOM } from 'jsdom';
import { registerBuiltInHelpers } from './helpers/built-in.js';
import { AfterAllPluginError, AfterTransformPluginError, ConfigError, ReadError, TransformPluginError, WriteError } from './utilities/errors.js';
import { isMarkdownFile, getDestinationFilename, getPathname } from './utilities/files.js';
import { parse as parseMarkdown, render as renderMarkdown } from './utilities/markdown.js';
import { render } from './utilities/render.js';
import { urlToPlainObject } from './utilities/urls.js';
export * from './types.js';
export const defaultConfig = {
    source: '.',
    destination: './_site',
    baseUrl: '',
    layouts: 'layouts',
    partials: 'partials',
    assets: '',
    data: {},
    cleanBeforePublishing: false,
    plugins: [],
    helpers: [],
    pageExtension: 'html',
    filterPages: () => true
};
export default class Doxicity {
    constructor(config) {
        this.hasPublished = false;
        this.hasRegisteredPartials = false;
        this.config = { ...defaultConfig, ...config };
        this.config.source = path.resolve(this.config.source);
        this.config.destination = path.resolve(this.config.destination);
        registerBuiltInHelpers(this.config);
        this.config.helpers.forEach(helper => {
            Handlebars.registerHelper(helper.name, helper.callback);
        });
    }
    async registerPartials() {
        const partialDir = path.resolve(this.config.source, this.config.partials);
        const partials = await globby([path.join(partialDir, '**/*.{hbs,md}')]);
        let partial = '';
        try {
            for await (partial of partials) {
                const parsed = path.parse(path.relative(partialDir, partial));
                const name = path.join(parsed.dir, parsed.name);
                const content = await fs.readFile(partial, 'utf-8');
                const html = renderMarkdown(content, { preserveHandlebars: true });
                Handlebars.registerPartial(name, html);
            }
        }
        catch (err) {
            const message = `Doxicity couldn't read the following partial from the file system.\n\n` +
                `${partial}\n\n` +
                `Please verify file permissions and try again.`;
            throw new ReadError(message);
        }
    }
    async renderPage(page) {
        return render(page, this.config);
    }
    static resolve(...args) {
        return path.resolve(...args);
    }
    async publish(options) {
        const opts = {
            copyAssets: true,
            ...options
        };
        const publishStartTime = Date.now();
        const ignoredDirectories = [
            path.join(this.config.source, this.config.layouts),
            path.join(this.config.source, this.config.partials)
        ];
        const publishedPages = [];
        if (!this.config.baseUrl || !/^https?:\/\//.test(this.config.baseUrl)) {
            const message = `An invalid baseUrl was detected in your configuration. Your baseUrl must be a URL that starts with http:// ` +
                `or https:// so Doxicity can correctly generate fully qualified URLs.`;
            throw new ConfigError(message);
        }
        const files = (await globby([
            path.resolve(this.config.source, '**/*.md'),
            ...ignoredDirectories.map(dir => `!${path.resolve(this.config.source, dir)}/**/*.md`)
        ])).filter(file => this.config.filterPages(file));
        if (!this.hasRegisteredPartials) {
            await this.registerPartials();
        }
        if (this.config.cleanBeforePublishing) {
            await this.clean();
        }
        for await (const file of files) {
            const sourceFile = file;
            const destinationFile = getDestinationFilename(sourceFile, this.config);
            const pathname = getPathname(sourceFile, this.config);
            const { frontMatter, content } = await parseMarkdown(sourceFile);
            const data = merge(this.config.data, frontMatter);
            const url = urlToPlainObject(new URL(pathname, this.config.baseUrl));
            Object.keys(data).forEach(key => {
                if (key.startsWith('$')) {
                    const message = `A field named "${key}" was found in your config data. Fields cannot start with $ because those names ` +
                        `are reserved for magic properties.`;
                    throw new ConfigError(message);
                }
            });
            data.$content = content;
            data.$url = url;
            const page = {
                sourceFile,
                destinationFile,
                pathname,
                data
            };
            publishedPages.push(page);
            let html = await this.renderPage(page);
            const doc = new JSDOM(html, {
                url: `https://internal/`
            }).window.document;
            for (const plugin of this.config.plugins) {
                if (plugin.transform) {
                    try {
                        await plugin.transform(doc, page);
                    }
                    catch (err) {
                        const message = `A plugin called "${plugin.name} threw an error during the "transform" phase of the following page.\n\n` +
                            `${page.sourceFile}\n\n` +
                            `Here's some more info about the error.\n\n` +
                            `${err.message}`;
                        throw new TransformPluginError(plugin.name, page, message);
                    }
                }
            }
            html = `<!DOCTYPE html>\n${doc.documentElement.outerHTML}`;
            for (const plugin of this.config.plugins) {
                if (plugin.afterTransform) {
                    try {
                        html = await plugin.afterTransform(html, page);
                    }
                    catch (err) {
                        const message = `A plugin called "${plugin.name} threw an error during the "afterTransform" phase of the following page.\n\n` +
                            `${page.sourceFile}\n\n` +
                            `Here's some more info about the error.\n\n` +
                            `${err.message}`;
                        throw new AfterTransformPluginError(plugin.name, page, message);
                    }
                    if (!html) {
                        const message = `A plugin called "${plugin.name} failed to return any HTML during the "afterTransform" phase of the following page.\n\n` +
                            `${page.sourceFile}\n\n` +
                            `Please update the plugin and try again.\n\n`;
                        throw new AfterTransformPluginError(plugin.name, page, message);
                    }
                }
            }
            try {
                await fs.mkdir(path.dirname(page.destinationFile), { recursive: true });
                await fs.writeFile(page.destinationFile, html, 'utf-8');
            }
            catch (err) {
                const message = `Doxicity couldn't write the following file to the file system.\n\n` +
                    `${page.destinationFile}\n\n` +
                    `Please verify that the target directory has the correct permissions and try again.`;
                throw new WriteError(message);
            }
        }
        const publishEndTime = Date.now();
        let assets = [];
        let timeToCopyAssets = 0;
        if (opts.copyAssets) {
            const result = await this.copyAssets();
            assets = result.assets;
            timeToCopyAssets = result.timeToCopy;
        }
        const publishResult = {
            pages: publishedPages,
            assets,
            timeToPublish: publishEndTime - publishStartTime,
            timeToCopyAssets
        };
        for await (const plugin of this.config.plugins) {
            if (plugin.afterAll) {
                try {
                    await plugin.afterAll(publishResult);
                }
                catch (err) {
                    const message = `A plugin called "${plugin.name} threw an error during the "afterAll" phase.\n\n` +
                        `Here's some more info about the error.\n\n` +
                        `${err.message}`;
                    throw new AfterAllPluginError(plugin.name, message);
                }
            }
        }
        return publishResult;
    }
    async copyAssets() {
        if (this.config.assets) {
            const startTime = Date.now();
            const from = `${path.join(this.config.source, this.config.assets)}/**/*`;
            const to = path.join(this.config.destination, this.config.assets);
            try {
                await fs.mkdir(to, { recursive: true });
                const assets = await cpy(from, to);
                const timeToCopy = Date.now() - startTime;
                return {
                    assets,
                    timeToCopy
                };
            }
            catch (err) {
                const message = `Doxicity wasn't able to copy assets from "${from}" to "${to}".\n\n` +
                    `Please verify that the target directories have the correct permissions and try again.`;
                throw new WriteError(message);
            }
        }
        return {
            assets: [],
            timeToCopy: 0
        };
    }
    async clean() {
        await fs.rm(this.config.destination, { recursive: true, force: true });
    }
    async serve(options) {
        const opts = {
            open: false,
            port: 8080,
            quiet: false,
            watch: false,
            onPublish: () => undefined,
            ...options
        };
        const output = (type, message) => {
            if (!opts.quiet) {
                console[type](message);
            }
        };
        if (!this.hasPublished) {
            const result = await this.publish();
            opts.onPublish(result);
        }
        this.bs = browserSync.create();
        this.bs.init({
            open: opts.open,
            port: opts.port,
            startPath: '/',
            logLevel: 'silent',
            logPrefix: '[doxicity]',
            logFileChanges: false,
            notify: false,
            ghostMode: false,
            server: {
                baseDir: this.config.destination
            }
        }, () => {
            const url = `http://localhost:${opts.port}/`;
            output('log', chalk.cyan(`\nLaunched the Doxicity dev server at ${url} 📗\n`));
            if (opts.watch) {
                output('log', `Watching for changes in: ${this.config.source}`);
                const pathsToWatch = path.join(this.config.source, '**/*');
                this.watcher = chokidar.watch(pathsToWatch, {
                    persistent: true,
                    ignored: [path.resolve(this.config.destination)],
                    ignoreInitial: true
                });
                this.watcher
                    .on('add', async (filename) => {
                    if (isMarkdownFile(filename)) {
                        output('log', `New page created: ${filename}`);
                        const result = await this.publish({ copyAssets: false });
                        opts.onPublish(result);
                    }
                    else {
                        output('log', `New file created: ${filename}`);
                        const result = await this.publish();
                        opts.onPublish(result);
                    }
                    this.bs.reload();
                })
                    .on('change', async (filename) => {
                    if (isMarkdownFile(filename)) {
                        output('log', `Page changed: ${filename}`);
                        const result = await this.publish({ copyAssets: false });
                        opts.onPublish(result);
                    }
                    else {
                        output('log', `File changed: ${filename}`);
                        const result = await this.publish();
                        opts.onPublish(result);
                    }
                    this.bs.reload();
                })
                    .on('unlink', async (filename) => {
                    output('log', `Page removed: ${filename}`);
                    try {
                        await this.clean();
                    }
                    catch (err) {
                        output('error', chalk.yellow(`Unable to clean the output directory before publishing: ${err.message}`));
                    }
                    try {
                        const result = await this.publish();
                        opts.onPublish(result);
                    }
                    catch (err) {
                        output('error', chalk.red(err.message));
                    }
                    this.bs.reload();
                });
            }
        });
    }
    async stop() {
        await Promise.all([this.bs.cleanup(), this.watcher.close()]);
    }
}
